from Crypto.Util import number
import os
import sys

"""
Die Methode generate_key() generiert die Schlüssel. Zuerst werden zwei 2048-Bit lange Zufalls-Primzahlen 
erzeugt, das entspricht Dezimalzahlen mit über 600 Stellen. Dabei wird ein Zufallsprimzahlengenerator aus 
dem Python Paket Crypto verwendet. Diese beiden grossen Primzahlen bilden den privaten Schlüssel.
Diese beiden Primzahlen werden anschliessend miteinander multipliziert, das erhaltene Produkt bildet den 
öffentlichen Schlüssel. Dieser öffentliche Schlüssel wird von der designierten Empfängerin der Nachricht 
versendet. Alle, welche Zugang zum öffentlichen Schlüssel haben, können damit eine Nachricht verschlüsseln. Es 
handelt sich dabei also um das geöffnete Schnappschloss, welches zum Verschliessen einer Botschaft verwendet 
werden kann. Die so verschlüsselte Nachricht kann jedoch nur mit Hilfe der beiden Primzahlen des privaten Schlüssels
wieder entschlüsselt werden. Es handelt sich dabei um den Schlüssel zum Schloss - deshalb müssen diese Primzahlen 
unbedingt sicher aufbewahrt werden. Die beiden Primzahlen des privaten Schlüssels werden deshalb auch Geheimzahlen
genannt. 
Weil es sich bei den Primzahlen um sehr grosse Zahlen handelt, ist ihr Produkt entsprechend auch eine sehr grosse Zahl. 
Für solch grosse Zahlen ist es auch mit den modernsten Computern und Verfahren praktisch unmöglich herauszufinden, aus 
welchen Faktoren sie zusammengesetzt sind - bei den in diesem Programm angewendeten Zahlen würde dies 
mehrere Milliarden Jahre dauern. Dies macht es möglich, dass das Produkt als öffentlicher 
Schlüssel versendet werden kann, ohne dass deswegen die beiden Geheimzahlen bekannt sind. Wer im Besitz des öffentlichen 
Schlüssels ist, kann deshalb die Nachricht zwar verschlüsseln, nicht aber entschlüsseln. Das ist die Grundlage des
Public Key Verfahrens. 
"""
def generate_key():
    p = number.getPrime(2048, os.urandom)
    q = number.getPrime(2048, os.urandom)
    n = p* q
    return p, q, n 

""" 
Für das Verschlüsseln der Nachricht wird neben der öffentlichen Zahl eine weitere Zahl benötigt, diese
muss im Verhältnis zur öffentlichen Zahl eine bestimmte Eigenschaft aufweisen, kann aber ansonsten
frei gewählt werden. In der Praxis wird oftmals mit der Zahl 65537 gearbeitet. Dass immer mit der gleichen
Zahl gearbeitet wird, hat keinen negativen Einfluss auf die Sicherheit des Verfahrens. In diesem Programm
wird auch mit dieser Zahl gearbeitet. Zum Entschlüsseln der Nachricht muss diese Zahl wiederum bekannt sein. 
"""
e = 65537

"""
In dieser Methode steckt das mathematische Element, welches das Public Key Verfahren nach RSA erst 
möglich macht. Der Basler Mathematiker Leonard Euler entdeckte im 18. Jahrhundert einen Zusammenhang
zwischen Zahlen, welcher der RSA-Entschlüsselung zugrunde liegt. Um diesen Zusammenhang anwenden zu 
können, muss für eine Zahl ihr Euler-Wert bestimmt werden. Der Euler-Wert x zu einer Zahl z 
ist die Anzahl Zahlen, welche kleiner als z sind und zu ihr teilerfremd sind. Wenn bekannt ist, dass es 
sich bei z um ein Produkt aus zwei Primzahlen p und q handelt, dann kann dieser Wert einfach mit der 
Formel x = (p - 1)(q -1) bestimmt werden. Wenn dies hingegen nicht bekannt ist, so ist es praktisch 
unmöglich, für eine grosse Zahl ihren Euler-Wert zu bestimmen. Diese Tatsache ist der Schlüssel zum RSA
Verfahren: Nur wer die beiden Geheimzahlen kennt, kann den Euler-Wert der öffentlichen Zahl bestimmen, 
da diese das Produkt der beiden Geheimzahlen ist. Zur Entschlüsselung der Nachricht wird 
aber der Euler-Wert der öffentlichen Zahl benötigt und kann somit nur von jenen ausgeführt werden, 
welche die Geheimzahlen kennen. Zur Verschlüsselung hingegen genügt es, die öffentliche Zahl zu kennen.
Die Methode euler(p,q) berechnet zu zwei übergebenen Zahlen den Euler-Wert des Produkts dieser beiden
Zahlen.
"""
def euler(p,q):
    x = (p - 1)*(q - 1)
    return x

"""
Diese Methode ist eine rekursive Umsetzung des sogenannten erweiterten Euklidschen Algorithmus. 
Dieser wird für die Methode modinv() verwendet.
"""
def egcd(a, b):
    if a == 0:
        return (b, 0, 1)
    else:
        g, y, x = egcd(b % a, a)
        return (g, x - (b // a) * y, y)

"""
Die Methode modinv(a,m) bestimmt zu den zwei übergebenen Zahlen a und m eine Zahl x, für welche
gilt, dass x*a mod m = 1 ist, das heisst, dass das Produkt von x und a dividiert durch m einen 
Rest von 1 ergibt. Diese Zahl x heisst multiplikative Inverse von a und m. Beispielsweise ist die 
multiplikative Inverse zu 2 und 5 die Zahl 3, da 2*3 dividiert durch 5 einen Rest von 1 ergibt. 
Im RSA-Verfahren wird die multiplikative Inverse zu einer gegebenen Zahl und dem Euler-Wert des 
öffentlichen Schlüssels zur Entschlüsselung der Nachricht benötigt. Da der Euler-Wert der Zahl nur
bestimmt werden kann, wenn die beiden Geheimzahlen bekannt sind, kann die multiplikative Inverse nur
bestimmen, wer die beiden Geheimzahlen kennt. 
"""
def modinv(a,m):
    g, x, y = egcd(a, m)
    return x % m

"""
Diese Methode wandelt eine übergebene Zahl in eine Binär-Zahl um und übergibt eine Liste, welche
die Ziffern gespeichert hat. Eine solche Liste wird für die Methode square_and_multiply(b,e,m) 
benötigt.
"""
def get_binary_digits(n):
    binary_digits = []
    while(n > 0):
        binary_digits.append(n%2)
        n = n//2
    binary_digits.reverse()
    return binary_digits

"""
Diese Methode berechnet den Rest, welcher bei der Division von b**e durch m entsteht. Für grosse
Werte von b und e handelt es sich dabei um eine schwierige Rechnung, da zuerst b**e berechnet 
werden muss und dies eine sehr grosse Zahl sein kann. Der Algorithmus hilft, indem er Erkenntnisse
aus der Zahlentheorie anwendet, so dass er die Berechnung von b**e vermeiden kann. Der Algorithmus
wird sowohl bei der Verschlüsselung als auch bei der Entschlüsselung angewendet. 
"""
def square_and_multiply(b,e,m):
    result = 1
    binary_e = get_binary_digits(e) 
    for i in range(len(binary_e)):
        result = (result * result) % m
        if(binary_e[i] == 1):
            result = (result * b) % m
    return result
  
"""
Diese Methode verschlüsselt die eingegebene Klartext-Nachricht mit dem öffentlichen Schlüssel und gibt die 
verschlüsselte Nachricht aus. Die Nachricht wird zuerst mit der Zahl e (in unserem Fall immer 65537) 
potenziert, anschliessend wird der Rest bestimmt, welcher bei der Division dieser Potenz durch die 
öffentliche Zahl entsteht. Für diese Operationen wird der square_and_multiply()-Algorithmus angewendet.
"""
def encrypt(klartext,n):
    return square_and_multiply(klartext,e,n)

"""
Diese Methode verschlüsselt die eingegebene Chiffre mit Hilfe des privaten Schlüssels. Dabei wird
die Chiffre mit dem aus den beiden Geheimzahlen bestimmten multiplikativen Inversen zur Euler-Wert
des öffentlichen Schlüssels potenziert, anschliessend wird der Rest bestimmt, welcher bei der Division
dieser Potenz durch den öffentlichen Schlüssel (welcher ja das Produkt der beiden Geheimzahlen ist) bestimmt.
Für dieser Operation wird der square_and_multiply()-Algorithmus angewendet. Ohne Kenntnis der beiden Geheimzahlen 
kann die Nachricht nicht entschlüsselt werden.
"""
def decrypt(chiffre,p,q):
    d = modinv(e,euler(p,q))
    return square_and_multiply(chiffre,d,p*q)


print("Wenn Sie einen Key generieren wollen, tippen Sie KEY.",
"\nWenn Sie eine Nachricht verschlüsseln wollen, tippen Sie DECRYPT.",
"\nWenn Sie eine Nachricht entschlüsseln wollen, tippen Sie ENCRYPT.")

action = input()


if(action == "KEY"):
    p, q, n = generate_key()
    print("Der öffentliche Schlüssel lautet:\n\n" + "n = ", n, 
    "\n\nDieser öffentliche Schlüssel ist das geöffnete Schloss, welches sie weitergeben können,", 
    "\ndamit an Sie gerichtete Nachrichten verschlüsselt werden können.",
    "\n\nDie beiden Geheimzahlen lauten:\n\n" + "p = ", p,
    "\n\nq = ", q,
    "\n\nDas ist ihr privater Schlüssel. Sie können damit die Nachrichten,", 
    "\nwelche mit dem von Ihnen versendeten öffentlichen Schlüssel verschlüsselt wurden,",
    "\nentschlüsseln. Speichern Sie diese beiden Zahlen an einem sicheren Ort.")


if(action == "ENCRYPT"):
    print("Geben Sie Ihre zu verschlüsselnde Nachricht ein:")
    klartext = int(input())
    print("Geben Sie den Public Key an:")
    public_key = int(input()) 
    print("Die verschlüsselte Nachricht lautet:\n", encrypt(klartext,public_key))


if(action == "DECRYPT"):
    print("Geben Sie die zu entschlüsselnde Nachricht ein:")
    chiffre = int(input())
    print("Geben Sie die erste der beiden Geheimzahlen ein:")
    private_key_1 = int(input())
    print("Geben Sie die zweite der beiden Geheimzahlen ein:")
    private_key_2 = int(input())
    print("Die entschlüsselte Nachricht lautet:\n", 
    decrypt(chiffre,private_key_1, private_key_2))
